@@ -41,7 +41,7 @@ our @EXPORT = qw(
 		 setup_forwarding
 		 );
 our @EXPORT_OK = qw( );
-our $VERSION = '4.4_6';
+our $VERSION = '4.4_7';
 
 #
 # ARP Filtering
@@ -130,7 +130,7 @@ sub setup_route_filtering() {
 	    emit   "fi\n";
 	}
 
-	if ( $capabilities{KERNELVERSION} < 20631 ) {
+	if ( have_capability( 'KERNELVERSION' ) < 20631 ) {
 	    emit 'echo 1 > /proc/sys/net/ipv4/conf/all/rp_filter';
 	} elsif ( $val ne '' ) {
 	    emit "echo $val > /proc/sys/net/ipv4/conf/all/rp_filter";
@@ -138,7 +138,7 @@ sub setup_route_filtering() {
 
 	emit "echo $val > /proc/sys/net/ipv4/conf/default/rp_filter" if $val ne '';
 
-	emit "[ -n \"\$NOROUTES\" ] || \$IP -4 route flush cache";
+	emit "[ -n \"\$g_noroutes\" ] || \$IP -4 route flush cache";
     }
 }
 
