@@ -11,7 +11,7 @@
 #       it under the terms of Version 2 of the GNU General Public License
 #       as published by the Free Software Foundation.
 #
-#       This program is distributed in the hope that it will be useful,
+#       This program is distributed in the shope that it will be useful,
 #       but WITHOUT ANY WARRANTY; without even the implied warranty of
 #       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 #       GNU General Public License for more details.
@@ -54,6 +54,7 @@ our @EXPORT = qw( NOTHING
 		  complex_zones
 		  non_firewall_zones
 		  single_interface
+		  chain_base
 		  validate_interfaces_file
 		  all_interfaces
 		  all_bridges
@@ -67,15 +68,19 @@ our @EXPORT = qw( NOTHING
 		  source_port_to_bridge
 		  interface_is_optional
 		  find_interfaces_by_option
+		  find_interfaces_by_option1
 		  get_interface_option
 		  set_interface_option
+		  verify_required_interfaces
+		  compile_updown
 		  validate_hosts_file
 		  find_hosts_by_option
 		  all_ipsets
+		  have_ipsec
 		 );
 
 our @EXPORT_OK = qw( initialize );
-our $VERSION = '4.4_6';
+our $VERSION = '4.4_10';
 
 #
 # IPSEC Option types
@@ -156,6 +161,7 @@ our @bport_zones;
 our %ipsets;
 our %physical;
 our $family;
+our $have_ipsec;
 
 use constant { FIREWALL => 1,
 	       IP       => 2,
@@ -178,9 +184,9 @@ use constant { SIMPLE_IF_OPTION   => 1,
 
 our %validinterfaceoptions;
 
-our %defaultinterfaceoptions = ( routefilter => 1 );
+our %defaultinterfaceoptions = ( routefilter => 1 , wait => 60 );
 
-our %maxoptionvalue = ( routefilter => 2, mss => 100000 );
+our %maxoptionvalue = ( routefilter => 2, mss => 100000 , wait => 120 );
 
 our %validhostoptions;
 
@@ -199,6 +205,7 @@ sub initialize( $ ) {
     @zones = ();
     %zones = ();
     $firewall_zone = '';
+    $have_ipsec = undef;
 
     @interfaces = ();
     %interfaces = ();
@@ -220,6 +227,7 @@ sub initialize( $ ) {
 				  nosmurfs    => SIMPLE_IF_OPTION + IF_OPTION_HOST,
 				  optional    => SIMPLE_IF_OPTION,
 				  proxyarp    => BINARY_IF_OPTION,
+				  required    => SIMPLE_IF_OPTION,
 				  routeback   => SIMPLE_IF_OPTION + IF_OPTION_ZONEONLY + IF_OPTION_HOST,
 				  routefilter => NUMERIC_IF_OPTION ,
 				  sourceroute => BINARY_IF_OPTION,
@@ -228,6 +236,7 @@ sub initialize( $ ) {
 				  upnpclient  => SIMPLE_IF_OPTION,
 				  mss         => NUMERIC_IF_OPTION,
 				  physical    => STRING_IF_OPTION + IF_OPTION_HOST,
+				  wait        => NUMERIC_IF_OPTION,
 				 );
 	%validhostoptions = (
 			     blacklist => 1,
@@ -245,15 +254,17 @@ sub initialize( $ ) {
 				    dhcp        => SIMPLE_IF_OPTION,
 				    maclist     => SIMPLE_IF_OPTION + IF_OPTION_HOST,
 				    nets        => IPLIST_IF_OPTION + IF_OPTION_ZONEONLY,
-				    nosmurfs    => SIMPLE_IF_OPTION,
+				    nosmurfs    => SIMPLE_IF_OPTION + IF_OPTION_HOST,
 				    optional    => SIMPLE_IF_OPTION,
 				    proxyndp    => BINARY_IF_OPTION,
+				    required    => SIMPLE_IF_OPTION,
 				    routeback   => SIMPLE_IF_OPTION + IF_OPTION_ZONEONLY + IF_OPTION_HOST,
 				    sourceroute => BINARY_IF_OPTION,
 				    tcpflags    => SIMPLE_IF_OPTION + IF_OPTION_HOST,
 				    mss         => NUMERIC_IF_OPTION,
 				    forward     => BINARY_IF_OPTION,
 				    physical    => STRING_IF_OPTION + IF_OPTION_HOST,
+				    wait        => NUMERIC_IF_OPTION,
 				 );
 	%validhostoptions = (
 			     blacklist => 1,
@@ -399,24 +410,19 @@ sub process_zone( \$ ) {
     }
 
     if ( $type eq IPSEC ) {
+	require_capability 'POLICY_MATCH' , 'IPSEC zones', '';
 	for ( @parents ) {
-	    unless ( $zones{$_}{type} == IPSEC ) {
-		set_super( $zones{$_} );
-	    }
+	    set_super( $zones{$_} ) unless $zones{$_}{type} == IPSEC;
 	}
     }
 
-    for ( $options, $in_options, $out_options ) {
-	$_ = '' if $_ eq '-';
-    }
-
     $zones{$zone} = { type       => $type,
 		      parents    => \@parents,
 		      bridge     => '',
 		      options    => { in_out  => parse_zone_option_list( $options || '', $type ) ,
 				      in      => parse_zone_option_list( $in_options || '', $type ) ,
 				      out     => parse_zone_option_list( $out_options || '', $type ) ,
-				      complex => ($type == IPSEC || $options || $in_options || $out_options ? 1 : 0) ,
+				      complex => ( $type == IPSEC || $options ne '-' || $in_options ne '-' || $out_options ne '-' ) ,
 				      nested  => @parents > 0 ,
 				      super   => 0 ,
 				    } ,
@@ -723,11 +729,30 @@ sub firewall_zone() {
     $firewall_zone;
 }
 
+#
+# Determine if the passed physical device is a bridge
+#
+sub is_a_bridge( $ ) {
+    which 'brctl' && qt( "brctl show | tail -n+2 | grep -q '^$_[0]\[\[:space:\]\]'" );
+} 
+
+#
+# Transform the passed interface name into a legal shell variable name.
+#
+sub chain_base($) {
+    my $chain = $_[0];
+
+    $chain =~ s/^@/at_/;
+    $chain =~ tr/[.\-%@]/_/;
+    $chain =~ s/\+$//;
+    $chain;
+}
+
 #
 # Process a record in the interfaces file
 #
-sub process_interface( $ ) {
-    my $nextinum = $_[0];
+sub process_interface( $$ ) {
+    my ( $nextinum, $export ) = @_;
     my $netsref = '';
     my ($zone, $originalinterface, $bcasts, $options ) = split_line 2, 4, 'interfaces file';
     my $zoneref;
@@ -742,9 +767,6 @@ sub process_interface( $ ) {
 	fatal_error "Firewall zone not allowed in ZONE column of interface record" if $zoneref->{type} == FIREWALL;
     }
 
-    $bcasts = '' if $bcasts eq '-';
-    $options  = '' if $options  eq '-';
-
     my ($interface, $port, $extra) = split /:/ , $originalinterface, 3;
 
     fatal_error "Invalid INTERFACE ($originalinterface)" if ! $interface || defined $extra;
@@ -752,7 +774,7 @@ sub process_interface( $ ) {
     if ( defined $port && $port ne '' ) {
 	fatal_error qq("Virtual" interfaces are not supported -- see http://www.shorewall.net/Shorewall_and_Aliased_Interfaces.html) if $port =~ /^\d+$/;
 	require_capability( 'PHYSDEV_MATCH', 'Bridge Ports', '');
-	fatal_error "Your iptables is not recent enough to support bridge ports" unless $capabilities{KLUDGEFREE};
+	fatal_error "Your iptables is not recent enough to support bridge ports" unless have_capability( 'KLUDGEFREE' );
 
 	fatal_error "Invalid Interface Name ($interface:$port)" unless $port =~ /^[\w.@%-]+\+?$/;
 	fatal_error "Duplicate Interface ($port)" if $interfaces{$port};
@@ -789,14 +811,14 @@ sub process_interface( $ ) {
     my $physical = $interface;
     my $broadcasts;
 
-    unless ( $bcasts eq '' || $bcasts eq 'detect' ) {
+    unless ( $bcasts eq '-' || $bcasts eq 'detect' ) {
 	my @broadcasts = split_list $bcasts, 'address';
 
 	for my $address ( @broadcasts ) {
 	    fatal_error 'Invalid BROADCAST address' unless $address =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/;
 	}
 
-	if ( $capabilities{ADDRTYPE} ) {
+	if ( have_capability( 'ADDRTYPE' ) ) {
 	    warning_message 'Shorewall no longer uses broadcast addresses in rule generation when Address Type Match is available';
 	} else {
 	    $broadcasts = \@broadcasts;
@@ -809,7 +831,9 @@ sub process_interface( $ ) {
 
     my $hostoptionsref = {};
 
-    if ( $options ) {
+    $options{ignore} = 1, $options = '-' if $options eq 'ignore';
+
+    if ( $options ne '-' ) {
 
 	my %hostoptions = ( dynamic => 0 );
 
@@ -913,22 +937,30 @@ sub process_interface( $ ) {
 	    }
 	}
 
+	fatal_error "Invalid combination of interface options" if $options{required} && $options{optional};
+
 	if ( $netsref eq 'dynamic' ) {
 	    my $ipset = "${zone}_" . chain_base $physical;
 	    $netsref = [ "+$ipset" ];
 	    $ipsets{$ipset} = 1;
 	}
 
-	$zoneref->{options}{in_out}{routeback} = 1 if $zoneref && $options{routeback};
-
 	if ( $options{bridge} ) {
 	    require_capability( 'PHYSDEV_MATCH', 'The "bridge" option', 's');
 	    fatal_error "Bridges may not have wildcard names" if $wildcard;
+	    $hostoptions{routeback} = $options{routeback} = 1;
 	}
 
-	$hostoptionsref = \%hostoptions;
+	$hostoptions{routeback} = $options{routeback} = is_a_bridge( $physical ) unless $export || $options{routeback};
 
-    }
+
+	$hostoptionsref = \%hostoptions;
+    } else {
+	#
+	# No options specified -- auto-detect bridge
+	#
+	$hostoptionsref->{routeback} = $options{routeback} = is_a_bridge( $physical ) unless $export;
+    }	
 
     $physical{$physical} = $interfaces{$interface} = { name       => $interface ,
 						       bridge     => $bridge ,
@@ -970,7 +1002,7 @@ sub validate_interfaces_file( $ ) {
 
     first_entry "$doing $fn...";
 
-    push @ifaces, process_interface( $nextinum++) while read_a_line;
+    push @ifaces, process_interface( $nextinum++, $export ) while read_a_line;
 
     #
     # We now assemble the @interfaces array such that bridge ports immediately precede their associated bridge
@@ -1013,7 +1045,7 @@ sub map_physical( $$ ) {
 #
 # Returns true if passed interface matches an entry in /etc/shorewall/interfaces
 #
-# If the passed name matches a wildcard, a entry for the name is added in %interfaces to speed up validation of other references to that name.
+# If the passed name matches a wildcard, an entry for the name is added in %interfaces to speed up validation of other references to that name.
 #
 sub known_interface($)
 {
@@ -1144,6 +1176,28 @@ sub find_interfaces_by_option( $ ) {
     \@ints;
 }
 
+#
+# Returns reference to array of interfaces with the passed option
+#
+sub find_interfaces_by_option1( $ ) {
+    my $option = $_[0];
+    my @ints = ();
+
+    for my $interface ( keys %interfaces ) {
+	my $interfaceref = $interfaces{$interface};
+
+	next unless defined $interfaceref->{physical};
+	next if $interfaceref->{physical} =~ /\+/;
+
+	my $optionsref = $interfaceref->{options};
+	if ( $optionsref && defined $optionsref->{$option} ) {
+	    push @ints , $interface
+	}
+    }
+
+    \@ints;
+}
+
 #
 # Return the value of an option for an interface
 #
@@ -1162,6 +1216,258 @@ sub set_interface_option( $$$ ) {
     $interfaces{$interface}{options}{$option} = $value;
 }
 
+#
+# Verify that all required interfaces are available after waiting for any that specify the 'wait' option.
+#
+sub verify_required_interfaces( $ ) {
+
+    my $generate_case = shift;
+    
+    my $returnvalue = 0;
+   
+    my $interfaces = find_interfaces_by_option 'wait';
+
+    if ( @$interfaces ) {
+	emit "local waittime\n";
+
+	for my $interface (@$interfaces ) {
+	    my $wait = $interfaces{$interface}{options}{wait};
+
+	    if ( $wait ) {
+		my $physical = get_physical $interface;
+	    
+		if ( $physical =~ /\+$/ ) {
+		    my $base = uc chain_base $physical;
+
+		    $physical =~ s/\+$/*/;
+
+		    emit( 'for interface in $(find_all_interfaces); do',
+			  '    case $interface in',
+			  "        $physical)",
+			  "            waittime=$wait",
+			  '            while [ $waittime -gt 0 ]; do',
+			  '                interface_is_usable $interface && break',
+			  '                waittime=$(($waittime - 1))',
+			  '            done',
+			  '            ;;',
+			  '    esac',
+			  'done',
+			  '',
+			);
+		} else {
+		    emit qq(if ! interface_is_usable $physical; then);
+		    emit qq(    waittime=$wait);
+		    emit  '';
+		    emit  q(    while [ $waittime -gt 0 ]; do);
+		    emit qq(        interface_is_usable $physical && break);
+		    emit  q(        sleep 1);
+		    emit   '        waittime=$(($waittime - 1))';
+		    emit  q(    done);
+		    emit qq(fi\n);
+		}
+
+		$returnvalue = 1;
+	    }
+	}
+    }
+
+    $interfaces = find_interfaces_by_option 'required';
+
+    if ( @$interfaces ) {
+
+	if ( $generate_case ) {
+	    emit( 'case "$COMMAND" in' );
+	    push_indent;
+	    emit( 'start|restart|restore|refresh)' );
+	    push_indent;
+	}
+
+	for my $interface (@$interfaces ) {
+	    my $physical = get_physical $interface;
+
+	    if ( $physical =~ /\+$/ ) {
+		my $base = uc chain_base $physical;
+
+		$physical =~ s/\+$/*/;
+
+		emit( "${base}_IS_UP=\n",
+		      'for interface in $(find_all_interfaces); do',
+		      '    case $interface in',
+		      "        $physical)",
+		      "            interface_is_usable \$interface && ${base}_IS_UP=Yes && break",
+		      '            ;;',
+		      '    esac',
+		      'done',
+		      '',
+		      "if [ -z \"\$${base}_IS_UP\" ]; then",
+		      "    startup_error \"None of the required interfaces $physical are available\"",
+		      "fi\n"
+		    );
+	    } else {
+		emit qq(if ! interface_is_usable $physical; then);
+		emit qq(    startup_error "Required interface $physical not available");
+		emit qq(fi\n);
+	    }
+	}
+	
+	if ( $generate_case ) {
+	    emit( ';;' );
+	    pop_indent;
+	    pop_indent;
+	    emit( 'esac' );
+	}
+
+	$returnvalue = 1;
+    }
+
+    $returnvalue;
+}
+
+#
+# Emit the updown() function
+#
+sub compile_updown() {
+    emit( '',
+	  '#',
+	  '# Handle the "up" and "down" commands',
+	  '#',
+	  'updown() # $1 = interface',
+	  '{',
+	);
+
+    push_indent;
+
+    emit( 'local state',
+	  'state=cleared',
+	  '' );
+
+    if ( $family == F_IPV4 ) {
+	emit 'if shorewall_is_started; then';
+    } else {
+	emit 'if shorewall6_is_started; then';
+    }
+
+    emit( '    state=started',
+	  'elif [ -f ${VARDIR}/state ]; then',
+	  '    case "$(cat ${VARDIR}/state)" in',
+	  '        Stopped*)',
+	  '            state=stopped',
+	  '            ;;',
+	  '        Cleared*)',
+	  '            ;;',
+	  '        *)',
+	  '            state=unknown',
+	  '            ;;',
+	  '    esac',
+	  'else',
+	  '    state=unknown',
+	  'fi',
+	  ''
+	);
+
+    emit( 'case $1 in' );
+
+    push_indent;
+
+    my $ignore   = find_interfaces_by_option 'ignore';
+    my $required = find_interfaces_by_option 'required';
+    my $optional = find_interfaces_by_option 'optional';
+
+    if ( @$ignore ) {
+	my $interfaces = join '|', map $interfaces{$_}->{physical}, @$ignore;
+
+	$interfaces =~ s/\+/*/;
+
+	emit( "$interfaces)",
+	      '    exit 0',
+	      '    ;;'
+	    );
+    }
+
+    if ( @$required ) {
+	my $interfaces = join '|', map $interfaces{$_}->{physical}, @$required;
+
+	my $wildcard = ( $interfaces =~ s/\+/*/ );
+
+	emit( "$interfaces)",
+	      '    if [ "$COMMAND" = up ]; then' );
+
+	if ( $wildcard ) {
+	    emit( '        if [ "$state" = started ]; then',
+		  '            COMMAND=restart',
+		  '        else',
+		  '            COMMAND=start',
+		  '        fi' );
+	} else {
+	    emit( '        COMMAND=start' );
+	}
+
+	emit( '        detect_configuration',
+	      '        define_firewall' );
+	
+	if ( $wildcard ) {
+	    emit( '    elif [ "$state" = started ]; then',
+		  '        COMMAND=restart',
+		  '        detect_configuration',
+		  '        define_firewall' );
+	} else {
+	    emit( '    else', 
+		  '        COMMAND=stop',
+		  '        detect_configuration',
+		  '        stop_firewall' );
+	}
+	
+	emit( '    fi',
+	      '    ;;'
+	    );
+    }
+
+    if ( @$optional ) {
+	my $interfaces = join '|', map $interfaces{$_}->{physical}, @$optional;
+
+	$interfaces =~ s/\+/*/;
+
+	emit( "$interfaces)",
+	      '    if [ "$COMMAND" = up ]; then',
+	      '        echo 0 > ${VARDIR}/${1}.state',
+	      '    else',
+	      '        echo 1 > ${VARDIR}/${1}.state',
+	      '    fi',
+	      '',
+	      '    if [ "$state" = started ]; then',
+	      '        COMMAND=restart',
+	      '        detect_configuration',
+	      '        define_firewall',
+	      '    elif [ "$state" = stopped ]; then',
+	      '        COMMAND=start',
+	      '        detect_configuration',
+	      '        define_firewall',
+	      '    fi',
+	      '    ;;',
+	    );
+    }
+
+    emit( "*)",
+	  '    case $state in',
+	  '        started)',
+	  '            COMMAND=restart',
+	  '            detect_configuration',
+	  '            define_firewall',
+	  '            ;;',
+	  '    esac',  
+	);
+
+    pop_indent;
+    
+    emit( 'esac' );
+
+    pop_indent;
+
+    emit( '}',
+	  '',
+	);
+}  
+
 #
 # Process a record in the hosts file
 #
@@ -1212,6 +1518,7 @@ sub process_host( ) {
 
 	for my $option ( @options ) {
 	    if ( $option eq 'ipsec' ) {
+		require_capability 'POLICY_MATCH' , q(The 'ipsec' option), 's';
 		$type = IPSEC;
 		$zoneref->{options}{complex} = 1;
 		$ipsec = 1;
@@ -1271,7 +1578,15 @@ sub validate_hosts_file()
 
     $ipsec |= process_host while read_a_line;
 
-    $capabilities{POLICY_MATCH} = '' unless $ipsec || haveipseczones;
+    $have_ipsec = $ipsec || haveipseczones;
+
+}
+
+#
+# Return an indication of whether IPSEC is present
+#
+sub have_ipsec() {
+    return defined $have_ipsec ? $have_ipsec : have_capability 'POLICY_MATCH';
 }
 
 #
